#![feature(const_trait_impl, const_ptr_as_ref, const_nonnull_new)]

// mod core_types;
pub mod fname;
pub mod global_tables;
pub mod tarray;
// pub mod types;
pub mod v2_types;

pub mod sdk {
    use std::{
        borrow::Cow,
        collections::{hash_map::Entry, BTreeMap, HashMap, HashSet},
        fmt::Display,
        sync::Mutex,
    };

    use anyhow::Context;

    use crate::{
        global_tables::objects::{FindClass, GOBJECTS},
        v2_types::{
            traits::{AsUObject, UEnumTrait, UObjectNonConst, UObjectTrait},
            UAnyType, UClass, UEnum, UObject,
        },
    };

    struct ClassCache {
        classes: Mutex<HashMap<String, UClass>>,
    }

    impl ClassCache {
        pub fn new() -> Self {
            Self {
                classes: Mutex::new(HashMap::new()),
            }
        }

        pub fn cache_class(&mut self, class: UClass) -> anyhow::Result<()> {
            let name = class.get_full_name()?;
            self.classes.lock().unwrap().insert(name, class);
            Ok(())
        }
    }

    impl FindClass for ClassCache {
        fn find_class<S>(&self, class_name: S) -> Option<UClass>
        where
            S: Into<String>,
        {
            let class_name = class_name.into();

            let class = self.classes.lock().unwrap().get(&class_name).cloned();

            match class {
                class @ Some(_) => class,
                None => match GOBJECTS
                    .read()
                    .unwrap()
                    .as_objects()
                    .unwrap()
                    .find_class(&class_name)
                {
                    Some(class) => {
                        self.classes
                            .lock()
                            .unwrap()
                            .insert(class_name, class.clone());
                        Some(class)
                    }
                    None => None,
                },
            }
        }
    }

    pub struct Sdk {}

    impl Sdk {
        pub fn package_objects() -> HashMap<UObject, Vec<UObject>> {
            let gobjects = GOBJECTS.read().unwrap();
            let objects = gobjects.as_objects().unwrap();

            let sorted_objects = objects
                .iter()
                // ensure item contains object
                .filter_map(|item| item.object())
                // get package object
                .map(|obj| (obj.package_object(), obj))
                .fold(
                    HashMap::<UObject, Vec<UObject>>::new(),
                    |mut acc, (pkg, obj)| {
                        match acc.entry(pkg) {
                            Entry::Occupied(mut entry) => {
                                entry.get_mut().push(obj);
                            }
                            Entry::Vacant(entry) => {
                                entry.insert(vec![obj]);
                            }
                        }

                        acc
                    },
                );

            sorted_objects
        }

        pub fn process_packages(packages: HashMap<UObject, Vec<UObject>>) -> anyhow::Result<()> {
            let mut cache = ClassCache::new();

            for (package, objects) in packages {
                Self::process_package(&mut cache, package, objects)?;
            }

            Ok(())
        }

        fn process_package(
            cache: &mut ClassCache,
            package: UObject,
            objects: Vec<UObject>,
        ) -> anyhow::Result<()> {
            cache
                .cache_class(package.class().context("package had no class")?)
                .context("package not of any base type")?;

            for object in objects {
                cache
                    .cache_class(object.class().expect("object had no class?"))
                    .context("object not of any base type?")?;

                let ty = UAnyType::from_uobject(object.as_uobject());
                log::debug!("ty: {ty:?}");

                match ty {
                    UAnyType::UObject(_) => {}
                    UAnyType::UClass(_) => {}
                    UAnyType::UField(_) => {}
                    UAnyType::UScriptStruct(_) => {}
                    UAnyType::UEnum(obj) => {
                        let enm = Self::process_enum(obj)?;
                        log::info!("enum: {enm}");
                    }
                    UAnyType::UStruct(_) => {}
                }
            }

            Ok(())
        }

        fn process_enum(enm: UEnum) -> anyhow::Result<Enum> {
            // get all the variants
            let values = enm
                .names()
                .iter()
                .enumerate()
                // canonicalize the names into valid rust
                .map(|(value, name)| {
                    let name = name.get_name().unwrap_or("AnonymousVariant".to_string());
                    let name = canonicalize_name(&name).to_string();
                    (value, name)
                })
                // store conflicts next to each other
                .fold(
                    HashMap::<String, Vec<usize>>::new(),
                    |mut acc, (value, name)| {
                        match acc.entry(name) {
                            Entry::Occupied(mut entry) => {
                                entry.get_mut().push(value);
                            }
                            Entry::Vacant(entry) => {
                                entry.insert(vec![value]);
                            }
                        }
                        acc
                    },
                );

            // sort by value
            let mut variants = BTreeMap::new();
            values.into_iter().for_each(|(name, values)| {
                if values.len() > 0 {
                    for (i, value) in values.into_iter().enumerate() {
                        variants.insert(value, format!("{name}{i}"));
                    }
                } else {
                    variants.insert(values.into_iter().next().unwrap(), name);
                }
            });

            let name = enm
                .as_uobject()
                .get_full_name()
                .context("enum name could not be found")?;
            let name = canonicalize_name(&name).to_string();

            Ok(Enum {
                name,
                values: variants.into_iter().map(|(_, name)| name).collect(),
            })
        }
    }

    fn keywords() -> HashSet<&'static str> {
        let mut keywords = HashSet::new();

        // rust keywords
        keywords.extend([
            "as", "break", "const", "continue", "crate", "else", "enum", "extern", "false", "fn",
            "for", "if", "impl", "in", "let", "loop", "match", "mod", "move", "mut", "pub", "ref",
            "return", "self", "Self", "static", "struct", "super", "trait", "true", "type",
            "unsafe", "use", "where", "while", "async", "await", "dyn", "abstract", "become",
            "box", "do", "final", "macro", "override", "priv", "typeof", "unsized", "virtual",
            "yield", "try",
        ]);

        // rust types
        keywords.extend([
            "bool", "f64", "f32", "str", "char", "u8", "u16", "u32", "u64", "u128", "i8", "i16",
            "i32", "i64", "i128", "usize", "isize",
        ]);

        keywords
    }

    fn token_chars() -> HashSet<char> {
        let mut chars = HashSet::new();

        chars.extend([
            ' ', '?', '+', '-', ':', '/', '^', '(', ')', '[', ']', '<', '>', '&', '.', '#', '\'',
            '"', '%',
        ]);

        chars
    }

    struct SplitResult<'a> {
        start: Option<&'a str>,
        middle: usize,
        end: Option<&'a str>,
    }

    impl<'a> SplitResult<'a> {
        pub fn is_valid(&self) -> bool {
            self.start.is_some() && self.middle == 0 && self.end.is_none()
        }

        pub fn into_valid(self) -> Cow<'a, str> {
            if self.is_valid() {
                Cow::Borrowed(self.start.unwrap())
            } else {
                let mut valid = self.start.map(|s| s.to_string()).unwrap_or_default();
                valid.extend(core::iter::repeat('_').take(self.middle));

                match self.end {
                    Some(end) => {
                        valid.push_str(&split_at_illegal_char(end).into_valid());
                    }
                    None => {}
                }

                Cow::Owned(valid)
            }
        }
    }

    fn empty_or_some(s: &str) -> Option<&str> {
        if s.is_empty() {
            None
        } else {
            Some(s)
        }
    }

    fn split_at_illegal_char(input: &str) -> SplitResult {
        let illegal_chars = token_chars();
        if let Some(pos) = input.find(|c| illegal_chars.contains(&c)) {
            let start = empty_or_some(&input[..pos]);
            // skip the illegal char
            let rest = &input[pos + 1..];

            if let Some(pos2) = rest.find(|c| !illegal_chars.contains(&c)) {
                SplitResult {
                    start,
                    middle: pos2 + 1,
                    end: empty_or_some(&rest[pos2..]),
                }
            } else {
                SplitResult {
                    start,
                    middle: 1,
                    end: empty_or_some(rest),
                }
            }
        } else {
            SplitResult {
                start: empty_or_some(input),
                middle: 0,
                end: None,
            }
        }
    }

    fn canonicalize_name(name: &str) -> Cow<str> {
        let valid = split_at_illegal_char(name).into_valid();
        if keywords().contains(valid.as_ref()) {
            Cow::Owned(format!("_{}", &valid))
        } else {
            valid
        }
    }

    #[derive(Debug)]
    struct Enum {
        name: String,
        values: Vec<String>,
    }

    impl Display for Enum {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            writeln!(f, "enum {} {{", self.name)?;
            for (i, name) in self.values.iter().enumerate() {
                writeln!(f, "\t{} = {},", name, i)?;
            }
            writeln!(f, "}}\n")
        }
    }
}
